use std::{ffi::CStr, os::fd::AsRawFd as _, process, time::Duration};

use chrono::Local;
use nix::{libc, sys::stat, unistd};
use once_cell::sync::Lazy;
use tokio::io::AsyncWriteExt as _;
use utmp_rs::{UtmpEntry, UtmpParser};

use crate::{paths, utmp};

// mimic upstream `getuidtty` function which caches running user and TTY names
static PW_NAME: Lazy<String> = Lazy::new(|| unsafe {
    let uid = unistd::getuid();

    let pwd = libc::getpwuid(uid.into());
    if !pwd.is_null() {
        CStr::from_ptr((*pwd).pw_name as *const _)
            .to_str()
            .unwrap()
            .to_owned()
    } else if !uid.is_root() {
        format!("uid {uid}")
    } else {
        String::from("root")
    }
});
static TTY_NAME: Lazy<String> = Lazy::new(|| unsafe {
    let ttyname = libc::ttyname(0);
    if !ttyname.is_null() {
        let mut tty_name = CStr::from_ptr(*ttyname as *const _).to_str().unwrap();
        debug!("retrieved TTY name: {}", tty_name);
        tty_name = tty_name.strip_prefix(paths::DEVPREFIX).unwrap_or(tty_name);
        tty_name = tty_name.strip_suffix('/').unwrap_or(tty_name);
        format!("({}) ", &tty_name[..utmp::UT_LINESIZE - 3])
    } else {
        warn!("could not retrieve TTY name");
        String::from("")
    }
});

// Check whether given filename looks like tty device.
fn file_isatty(fname: &str) -> bool {
    match stat::stat(fname) {
        Err(error) => {
            warn!("could not stat {}: {}", fname, error);
            false
        }
        Ok(st) => {
            if st.st_nlink != 1 || (st.st_mode & libc::S_IFCHR) != libc::S_IFCHR {
                return false;
            }

            /*
             *  It would be an impossible task to list all major/minors
             *  of tty devices here, so we just exclude the obvious
             *  majors of which just opening has side-effects:
             *  printers and tapes.
             */
            ![1, 2, 6, 9, 12, 16, 21, 27, 37, 96, 97, 206, 230].contains(&stat::major(st.st_dev))
        }
    }
}

// Wall function.
pub fn wall(text: &str, remote: bool) {
    // Get and report current hostname, to make it easier to find
    // out which machine is being shut down.
    let hostname = match unistd::gethostname() {
        Ok(hostname) => {
            debug!("retrieved hostname: {:?}", hostname);
            hostname.to_str().unwrap().to_string()
        }
        Err(error) => {
            warn!("could not retrieve hostname: {}", error);
            String::from("[unknown]")
        }
    };

    // Get the time
    let date = Local::now().format("%a %b %d %H:%M:%S %Y");

    let broadcast_message = if !remote {
        format!("\r\nRemote broadcast message ({date}):\r\n\r\n")
    } else {
        format!(
            "\r\nBroadcast message from {}@{hostname} {}({date}):\r\n\r\n",
            *PW_NAME, *TTY_NAME
        )
    };

    // Fork to avoid us hanging in a write()
    match unsafe { unistd::fork() } {
        Ok(unistd::ForkResult::Parent { child, .. }) => {
            debug!("successfully forked, child has pid: {}", child)
        }
        Ok(unistd::ForkResult::Child) => {
            wall_utmp_users(text, &broadcast_message);
            process::exit(0);
        }
        Err(error) => warn!("could not fork: {}", error),
    }
}

fn wall_utmp_users(text: &str, broadcast_message: &str) {
    let utmp_parser = UtmpParser::from_path(paths::UTMP);
    if let Err(error) = utmp_parser {
        warn!("could not open {}: {}", paths::UTMP, error);
        return;
    }

    let rt = tokio::runtime::Builder::new_current_thread()
        .enable_time()
        .build()
        .unwrap();

    for entry in utmp_parser.unwrap() {
        if let Err(error) = entry {
            warn!("could not parse utmp entry: {}", error);
            continue;
        }

        // is this an `UserProcess` utmp entry ?
        if let UtmpEntry::UserProcess { line, user, .. } = entry.unwrap() {
            if user.is_empty() {
                continue;
            }

            let term = if !line.starts_with(paths::DEVPREFIX) {
                paths::DEVPREFIX.to_owned() + &line[..utmp::UT_LINESIZE]
            } else {
                line[..utmp::UT_LINESIZE - paths::DEVPREFIX.len()].to_string()
            };

            if term.contains("/../") {
                continue;
            }

            // Open it non-delay
            if !file_isatty(&term) {
                continue;
            }
            let message = broadcast_message.to_owned() + &text.escape_default().to_string();
            rt.block_on(wall_user(&term, &message));
        }
    }
}

async fn wall_user(term: &str, message: &str) {
    if tokio::time::timeout(Duration::from_secs(2), async {
        match tokio::fs::File::options()
            .write(true)
            .custom_flags(libc::O_NDELAY | libc::O_NOCTTY)
            .open(&term)
            .await
        {
            Err(error) => {
                warn!("could not open {}: {}", term, error);
            }
            Ok(mut file) => match unistd::isatty(file.as_raw_fd()) {
                Err(error) => {
                    warn!(
                        "could not check whether {} refers to a terminal: {}",
                        term, error
                    );
                }
                Ok(is_a_tty) => {
                    if !is_a_tty {
                        debug!("{} is not a terminal, skipping wall...", term);
                        return;
                    }

                    if let Err(error) = file.write_all(message.as_bytes()).await {
                        warn!("could not write wall message to {}: {}", term, error);
                        return;
                    }

                    if let Err(error) = file.flush().await {
                        warn!("could not flush wall message to {}: {}", term, error);
                    }
                }
            },
        }
    })
    .await
    .is_err()
    {
        warn!("timeout writing wall message to {}", term);
    }
}
