use std::{
    env,
    fs::{self, File, Permissions},
    io::{self, BufRead as _, ErrorKind, Write as _},
    iter,
    os::{
        fd::AsRawFd as _,
        unix::{
            fs::{MetadataExt as _, PermissionsExt as _},
            process::CommandExt as _,
        },
    },
    path::Path,
    process, thread,
    time::Duration,
};

use chrono::{Duration as ChronoDuration, Local, NaiveTime, Timelike as _};
use clap::{Command, Parser};
use env_logger::{Builder, DEFAULT_FILTER_ENV, DEFAULT_WRITE_STYLE_ENV};
use log::LevelFilter;
use nix::{
    errno::Errno,
    libc,
    sys::{
        reboot::{self, RebootMode},
        signal::{self, SigAction, SigHandler, Signal},
        stat::{self, Mode},
    },
    unistd::{self, Pid},
};
use syslog::{self, Formatter3164};
use tokio::{self, io::AsyncWriteExt as _};
use utmp_rs::{UtmpEntry, UtmpParser};

mod dowall;
mod initreq;
mod paths;
mod util;
mod utmp;

#[macro_use]
extern crate log;

// default time to wait between TERM and KILL
const WAIT_BETWEEN_SIGNALS: u32 = 3;

// Break off an already running shutdown.
extern "C" fn stop_it(signal: libc::c_int) {
    if let Ok(signal) = Signal::try_from(signal) {
        debug!("caught signal {}, cancelling shutdown...", signal);
    }

    for file in &[
        paths::NOLOGIN,
        paths::FASTBOOT,
        paths::FORCEFSCK,
        paths::SDPID,
    ] {
        if let Err(error) = fs::remove_file(file) {
            if error.kind() != ErrorKind::NotFound {
                warn!("could not remove {}: {}", file, error);
            }
        }
    }

    println!("\r\nShutdown cancelled.");
    process::exit(0);
}

// Read SDALLOW file and walk through UTMP file to find any logged in allowed user.
fn is_user_allowed() -> Result<bool, io::Error> {
    let file = File::open(paths::SDALLOW)?;

    // check file permissions
    match stat::fstat(file.as_raw_fd()) {
        Err(error) => {
            error!("fstat: {}", error);
            process::exit(1);
        }
        Ok(file_stat) => {
            if file_stat.st_uid != 0
                || !(Mode::from_bits_truncate(file_stat.st_mode) & (Mode::S_IWGRP | Mode::S_IWOTH))
                    .is_empty()
            {
                warn!(
                    "too broad permissions for {}: {:?}",
                    paths::SDALLOW,
                    Mode::from_bits_truncate(file_stat.st_mode)
                );
                return Ok(false);
            }
        }
    }

    // parse authorized users
    let mut allowed_users: Vec<String> = vec![
        // root is always OK.
        String::from("root"),
    ];
    for line in io::BufReader::new(file).lines() {
        if let Err(error) = line {
            warn!("could not read line: {}", error);
            continue;
        }

        // trim whitespaces before processing line
        let read_line = line?.split_whitespace().collect::<Vec<&str>>().join(" ");
        if read_line.starts_with('#') {
            debug!("ignored comment line: {}", read_line);
            continue;
        }

        debug!("found allowed user: {}", read_line);
        allowed_users.push(read_line.clone());
    }

    // Now walk through /var/run/utmp to find logged in users.
    for entry in UtmpParser::from_path(paths::UTMP)? {
        if let Err(error) = entry {
            warn!("could not parse utmp entry: {}", error);
            continue;
        }

        // is this an `UserProcess` utmp entry ?
        if let UtmpEntry::UserProcess { line, user, .. } = entry.unwrap() {
            // check TTY device identifier
            let mut user_tty = paths::DEVPREFIX.to_owned();
            user_tty.push_str(&line);
            match fs::metadata(&user_tty) {
                Err(error) => {
                    if error.kind() != ErrorKind::NotFound {
                        warn!("could not query {} file metadata: {}", user_tty, error);
                    }
                    continue;
                }
                Ok(metadata) => {
                    if stat::major(metadata.dev()) == 4 && stat::minor(metadata.dev()) < 64 {
                        // is this user allowed ?
                        if allowed_users.contains(&user) {
                            return Ok(true);
                        }
                    }
                }
            }
        }
    }

    Ok(false)
}

// Decode shutdown time.
fn time_to_duration(time_string: &str) -> Duration {
    // alias "now"
    if time_string == "now" {
        return Duration::ZERO;
    }

    // validate time argument
    if time_string
        .chars()
        .any(|c| c != '+' && c != ':' && !c.is_ascii_digit())
    {
        error!("invalid time string: {}", time_string);
        process::exit(2);
    }

    // take care of old '+mins' notation
    let is_relative = time_string.starts_with('+');
    let time = if is_relative {
        &time_string[1..]
    } else {
        time_string
    };

    if !time.contains(':') {
        // time in minutes
        match time.parse::<u32>() {
            Err(error) => {
                error!("could not parse time: {}", error);
                process::exit(2);
            }
            Ok(minutes) => return Duration::from_secs((minutes * 60).into()),
        }
    }

    // time in HH:MM format
    match NaiveTime::parse_from_str(time, "%H:%M") {
        Err(error) => {
            error!("could not parse time string: {}", error);
            process::exit(2);
        }
        Ok(time) => {
            debug!("successfully parsed time string: {:?}", time);
            let time_seconds = time.hour() * 3600 + time.minute() * 60;
            if is_relative {
                // hours and minutes from now
                Duration::from_secs(time_seconds.into())
            } else {
                // time of day
                let local_time = Local::now().time();

                let local_time_seconds =
                    local_time.hour() * 3600 + local_time.minute() * 60 + local_time.second();

                // if specified time corresponds to a relatively "sooner" time today, we add
                // 24h to point "tomorrow"
                Duration::from_secs(
                    (time_seconds.abs_diff(local_time_seconds)
                        + if time_seconds < local_time_seconds {
                            24 * 3600
                        } else {
                            0
                        })
                    .into(),
                )
            }
        }
    }
}

// Set environment variables in the init process.
fn init_setenv(name: &str, value: Option<&str>) -> bool {
    let mut env_var = name.to_owned();
    if let Some(v) = value {
        env_var.push('=');
        env_var.push_str(v);
    }
    env_var.push('\x00');

    if env_var.len() > initreq::DATA_LENGTH {
        warn!(
            "init_setenv: environment variable too long ({} bytes)",
            env_var.len()
        );
        return false;
    }

    let init_request = initreq::InitRequest {
        magic: initreq::INIT_MAGIC,
        cmd: initreq::INIT_CMD_SETENV,
        runlevel: 0,
        sleeptime: 0,
        data: &env_var
            .bytes()
            .chain(iter::repeat(b'\x00'))
            .take(initreq::DATA_LENGTH)
            .collect::<Vec<_>>()
            .try_into()
            .unwrap(),
    };

    // Open the fifo and write the command.
    // Make sure we don't hang on opening /run/initctl
    tokio::runtime::Builder::new_current_thread()
        .enable_time()
        .build()
        .unwrap()
        .block_on(async {
            match tokio::time::timeout(Duration::from_secs(3), async {
                match tokio::fs::File::options()
                    .write(true)
                    .open(initreq::INIT_FIFO)
                    .await
                {
                    Err(error) => {
                        eprintln!("shutdown: {}: {}\n", initreq::INIT_FIFO, error);
                        false
                    }
                    Ok(mut file) => {
                        for byte in &bincode::serialize(&init_request).unwrap() {
                            if let Err(error) = file.write_u8(*byte).await {
                                if error.kind() == ErrorKind::Interrupted
                                    || error.kind() == ErrorKind::WouldBlock
                                {
                                    continue;
                                }

                                break;
                            }
                        }

                        true
                    }
                }
            })
            .await
            {
                Err(_) => {
                    eprintln!(
                        "shutdown: timeout opening/writing control channel {}\n",
                        initreq::INIT_FIFO
                    );
                    false
                }
                Ok(is_ok) => {
                    debug!(
                        "successfully set {} init process environment variable",
                        env_var
                    );
                    is_ok
                }
            }
        })
}

// Tell everyone the system is going down in [wait_duration] minutes.
fn issue_warn(wait_duration: Duration, warn_message: &str, new_state: &str) {
    let mut message = warn_message.to_string();
    let minutes = wait_duration.as_secs() % 60;
    if minutes == 0 {
        message.push_str(&format!("\rThe system is going down {new_state} NOW!\r\n"));
    } else {
        message.push_str(&format!(
            "\rThe system is going down {new_state} in {minutes} minute{}!\r\n",
            if minutes > 1 { "s" } else { "" }
        ));
    }

    dowall::wall(&message, false);
}

// Kill all processes, call /etc/init.d/halt (if present).
fn fastdown(down_level: char, warn_delay: Option<u32>) -> ! {
    // first close all files
    for std_stream_fd in [
        std::io::stdin().as_raw_fd(),
        std::io::stdout().as_raw_fd(),
        std::io::stderr().as_raw_fd(),
    ] {
        match unistd::isatty(std_stream_fd) {
            Err(error) => {
                warn!(
                    "could not check whether {} refers to a terminal: {}",
                    std_stream_fd, error
                );
                continue;
            }
            Ok(is_a_tty) => {
                if is_a_tty {
                    continue;
                }

                match unistd::close(std_stream_fd) {
                    Err(error) => warn!(
                        "could not close standard stream {}: {}",
                        std_stream_fd, error
                    ),
                    Ok(()) => {
                        if let Err(error) = File::open(paths::DEVNULL) {
                            warn!(
                                "could not open {} as {}: {}",
                                paths::DEVNULL,
                                std_stream_fd,
                                error
                            );
                        }
                    }
                }
            }
        }
    }

    let mut fds: Vec<i32> = (3..20).collect();
    fds.push(255);
    for fd in fds {
        if let Err(error) = unistd::close(fd) {
            if error != Errno::EBADF {
                warn!("could not close file descriptor {}: {}", fd, error);
            }
        }
    }

    // idle init
    if let Err(error) = signal::kill(Pid::from_raw(1), Signal::SIGTSTP) {
        eprintln!("shutdown: can't idle init: {}.", error);
        process::exit(1);
    }

    // kill all processes
    eprintln!("shutdown: sending all processes the TERM signal...");
    if let Err(error) = signal::kill(Pid::from_raw(-1), Signal::SIGTERM) {
        warn!("could not send TERM signal to all processes: {}", error);
    }
    thread::sleep(Duration::from_secs(
        warn_delay.unwrap_or(WAIT_BETWEEN_SIGNALS).into(),
    ));
    eprintln!("shutdown: sending all processes the KILL signal.");
    if let Err(error) = signal::kill(Pid::from_raw(-1), Signal::SIGKILL) {
        warn!("could not send KILL signal to all processes: {}", error);
    }

    // record the fact that we're going down
    utmp::write_wtmp(
        "shutdown",
        "~~",
        Pid::from_raw(0),
        utmp::RecordType::RunLvl,
        "~~",
    );

    // this is for those who have quota installed
    if let Err(error) = unistd::acct::disable() {
        eprintln!("shutdown: can not stop process accounting: {}.", error);
    }

    if let Err(error) = process::Command::new("quotaoff")
        .arg("-a")
        .env_remove(DEFAULT_FILTER_ENV)
        .env_remove(DEFAULT_WRITE_STYLE_ENV)
        .spawn()
    {
        warn!("could not spawn quotaoff: {}", error);
    }

    unistd::sync();

    eprintln!("shutdown: turning off swap");
    if let Err(error) = process::Command::new("swapoff")
        .arg("-a")
        .env_remove(DEFAULT_FILTER_ENV)
        .env_remove(DEFAULT_WRITE_STYLE_ENV)
        .spawn()
    {
        warn!("could not spawn swapoff: {}", error);
    }

    eprintln!("shutdown: unmounting all file systems");
    if let Err(error) = process::Command::new("umount")
        .arg("-a")
        .env_remove(DEFAULT_FILTER_ENV)
        .env_remove(DEFAULT_WRITE_STYLE_ENV)
        .spawn()
    {
        warn!("could not spawn umount: {}", error);
    }

    // we're done, halt or reboot now
    let do_halt = down_level == '0';
    eprintln!(
        "{}",
        if do_halt {
            "The system is halted. Press CTRL-ALT-DEL or turn off power."
        } else {
            "Please stand by while rebooting the system."
        }
    );
    if let Err(error) = reboot::reboot(if do_halt {
        RebootMode::RB_HALT_SYSTEM
    } else {
        RebootMode::RB_AUTOBOOT
    }) {
        error!("could not reboot: {}", error);
        process::exit(1);
    }

    process::exit(0);
}

// Go to runlevel 0, 1 or 6.
fn issue_shutdown(
    halt_type: Option<&str>,
    down_level: char,
    dont_shut: bool,
    no_init: bool,
    warn_delay: Option<u32>,
    warn_message: &str,
    new_state: &str,
) -> ! {
    // warn for the last time
    issue_warn(Duration::ZERO, warn_message, new_state);

    if dont_shut {
        thread::sleep(Duration::from_secs(1));
        stop_it(0);
    }

    let formatter = Formatter3164 {
        process: "shutdown".into(),
        ..Formatter3164::default()
    };

    match syslog::unix(formatter.clone()) {
        Err(error) => warn!("could not connect to syslog: {}", error),
        Ok(mut logger) => {
            if let Err(error) = logger.notice(&format!(
                "shutting down for system {}",
                if down_level == '0' { "halt" } else { "reboot" }
            )) {
                warn!("could not write to syslog: {}", error);
            }
        }
    }

    // see if we have to do it ourselves
    if no_init {
        fastdown(down_level, warn_delay);
    }

    if let Err(error) = fs::remove_file(paths::SDPID) {
        if error.kind() != ErrorKind::NotFound {
            warn!("could not remove {}: {}", paths::SDPID, error);
        }
    }
    if let Err(error) = fs::remove_file(paths::NOLOGIN) {
        if error.kind() != ErrorKind::NotFound {
            warn!("could not remove {}: {}", paths::NOLOGIN, error);
        }
    }

    // create the arguments for init
    let mut command = process::Command::new(paths::INIT);
    if let Some(warn_delay) = warn_delay {
        command.args(["-t", &warn_delay.to_string()]);
    }
    command
        .arg(down_level.to_string())
        .env_remove(DEFAULT_FILTER_ENV)
        .env_remove(DEFAULT_WRITE_STYLE_ENV);

    // now execute init to change runlevel
    unistd::sync();
    init_setenv("INIT_HALT", halt_type);
    let error = command.exec();

    // oops - failed
    eprintln!("\rshutdown: cannot execute {}: {}", paths::INIT, error);
    if let Err(error) = fs::remove_file(paths::FASTBOOT) {
        if error.kind() != ErrorKind::NotFound {
            warn!("could not remove {}: {}", paths::FASTBOOT, error);
        }
    }
    if let Err(error) = fs::remove_file(paths::FORCEFSCK) {
        if error.kind() != ErrorKind::NotFound {
            warn!("could not remove {}: {}", paths::FORCEFSCK, error);
        }
    }

    init_setenv("INIT_HALT", None);

    match syslog::unix(formatter) {
        Err(error) => warn!("could not connect to syslog: {}", error),
        Ok(mut logger) => {
            if let Err(error) = logger.notice("shutdown failed") {
                warn!("could not write to syslog: {}", error);
            }
        }
    }
    process::exit(1);
}

// Returns if a warning is to be sent for [wait_duration].
fn need_warning(wait_duration: Duration, quiet_full: bool, quiet_partial: bool) -> bool {
    if quiet_full {
        return false;
    }

    let minutes = wait_duration.as_secs() / 60;
    if quiet_partial {
        return minutes == 10 || minutes == 5 || minutes % 60 == 0;
    }

    // no silence setting, print lots of warnings
    minutes < 10 || {
        if minutes < 60 {
            minutes % 15 == 0
        } else if minutes < 180 {
            minutes % 30 == 0
        } else {
            minutes % 60 == 0
        }
    }
}

#[derive(Parser, Debug)]
#[clap(version, about)]
struct Args {
    /// enable informational logs
    #[clap(short, long)]
    verbose: bool,

    /// enable debugging logs
    #[clap(short, long)]
    debug: bool,

    /// use /etc/shutdown.allow
    #[clap(short = 'a')]
    use_acl: bool,

    /// don't really shutdown, only warn
    #[clap(short = 'k')]
    dont_shut: bool,

    /// reboot after shutdown
    #[clap(short = 'r')]
    reboot: bool,

    /// halt after shutdown
    #[clap(short = 'h')]
    halt: bool,

    /// halt action is to turn off power, can only be used with along with -h flag
    #[clap(short = 'P')]
    halt_type_power: bool,

    /// halt action is to just halt, can only be used with along with -h flag
    #[clap(short = 'H')]
    halt_type_halt: bool,

    /// do a "fast" boot (skip fsck)
    #[clap(short)]
    fast_boot: bool,

    /// force fsck on boot
    #[clap(short = 'F')]
    force_fsck: bool,

    /// do not go through "init" but go down real fast
    #[clap(short = 'n')]
    no_init: bool,

    /// cancel a running shutdown
    #[clap(short)]
    cancel: bool,

    /// quiet mode - display fewer shutdown warnings
    #[clap(short = 'q')]
    quiet_partial: bool,

    /// full quiet mode - display only final shutdown warning
    #[clap(short = 'Q')]
    quiet_full: bool,

    /// delay between warning message and kill signal (seconds)
    #[clap(short = 't')]
    warn_delay: Option<u32>,

    /// ignored for System-V compatibility
    #[clap(short = 'y')]
    _ignored: bool,

    /// System-V style to specify time
    #[clap(short = 'g')]
    sysv_time: Option<String>,

    /// level to go to (could be '0', '1', '5', '6', 'A', 'B', 'C' or 'S')
    #[clap(short = 'i')]
    down_level: Option<char>,

    // Shutdown time (try "now")
    time: Option<String>,

    // Shutdown warning message
    #[clap(takes_value = true, multiple_values = true, last = true)]
    warn_message: Vec<String>,
}

/*
 *  Main program.
 *  Process the options and do the final countdown.
 */
fn main() {
    let args = Args::parse();

    let mut builder = Builder::new();
    builder.filter_level(if args.debug {
        LevelFilter::Debug
    } else if args.verbose {
        LevelFilter::Info
    } else {
        LevelFilter::Warn
    });
    builder.parse_default_env();
    builder.init();

    // We can be installed setuid root (executable for a special group).
    // This way is risky, do error check on setuid call.
    if let Err(error) = unistd::setuid(unistd::geteuid()) {
        error!("setuid: {}", error);
        process::abort();
    }

    if !unistd::getuid().is_root() {
        eprintln!("shutdown: you must be root to do that!");
        process::exit(1);
    }

    let halt_type: Option<&str> = if args.halt_type_halt {
        Some("HALT")
    } else if args.halt_type_power {
        Some("POWER")
    } else {
        None
    };

    let down_level: char = if args.reboot {
        '6'
    } else if args.halt {
        '0'
    } else {
        match args.down_level {
            Some(down_level) => {
                if "0156aAbBcCsS".chars().all(|c| c != down_level) {
                    error!("shutdown: `{}': bad runlevel", down_level);
                    process::exit(2);
                }

                down_level
            }
            None => '1',
        }
    };

    if halt_type.is_some() && down_level != '0' {
        eprintln!("shutdown: -H and -P flags can only be used along with -h flag.");
        process::exit(2);
    }

    // do we need to use the shutdown.allow file ?
    if args.use_acl {
        match is_user_allowed() {
            Err(error) => {
                if error.kind() != ErrorKind::NotFound {
                    error!("could not check authorized users: {}", error);
                    process::exit(1);
                }

                warn!("could not process logged in users: {}", error);
            }
            Ok(is_allowed) => {
                if !is_allowed {
                    if let Ok(mut file) = File::open(paths::CONSOLE) {
                        if let Err(error) = file
                            .write_all("\rshutdown: no authorized users logged in.\r\n".as_bytes())
                        {
                            warn!("could not write permission error to console: {}", error);
                        }
                    }
                    process::exit(1);
                }

                debug!("user is allowed by system!");
            }
        }
    }

    // read running shutdown PID from file (if any)
    let mut pid = Pid::from_raw(0);
    if let Ok(file) = File::open(paths::SDPID) {
        let mut buf = String::new();
        if io::BufReader::new(file).read_line(&mut buf).is_ok() {
            match buf.trim().parse::<i32>() {
                Err(error) => warn!("could not read PID from file: {}", error),
                Ok(read_pid) => {
                    pid = Pid::from_raw(read_pid);
                    debug!("found running shutdown with PID={}", pid);
                }
            }
        }
    }

    // "read remaining words"
    let mut warn_message = String::new();
    for string in &args.warn_message {
        if !warn_message.is_empty() {
            warn_message.push(' ');
        }
        warn_message.push_str(string);
    }
    if !warn_message.is_empty() {
        warn_message.push_str("\r\n");
    }

    // see if we want to run or cancel
    if args.cancel {
        if pid.as_raw() <= 0 {
            eprintln!("shutdown: cannot find pid of running shutdown.");
            process::exit(1);
        }

        init_setenv("INIT_HALT", None);

        if signal::kill(pid, Signal::SIGINT).is_err() {
            eprintln!("shutdown: not running.");
            process::exit(1);
        }

        if !warn_message.is_empty() {
            dowall::wall(&warn_message, false);
        }

        process::exit(0);
    }

    // check syntax
    let time = match args.time {
        Some(time) => time,
        None => {
            if let Some(time) = args.sysv_time {
                time
            } else {
                println!("{}", Command::new("shutdownr").render_usage());
                process::exit(2);
            }
        }
    };

    // see if we are already running
    if pid.as_raw() > 0 && signal::kill(pid, None).is_ok() {
        eprintln!("\rshutdown: already running.");
        process::exit(1);
    }

    // extra check
    if args.no_init && down_level != '0' && down_level != '6' {
        eprintln!("shutdown: can use \"-n\" for halt or reboot only.");
        process::exit(1);
    }

    // tell users what we're gonna do
    let to_run_level = format!("to run level {down_level}");
    let new_state: &str = match down_level {
        '0' => "for system halt",
        '6' => "for reboot",
        '1' => "to maintenance mode",
        _ => &to_run_level,
    };

    // go the root directory
    if let Err(error) = env::set_current_dir(Path::new("/")) {
        eprintln!("chdir(/): {}", error);
        process::exit(1);
    }

    // create a new PID file
    if let Err(error) = fs::remove_file(paths::SDPID) {
        if error.kind() != ErrorKind::NotFound {
            error!("could not remove shutdown PID file: {}", error);
            process::exit(1);
        }
    }
    match File::create(paths::SDPID) {
        Err(error) => {
            // `ErrorKind::ReadOnlyFilesystem` is a nightly-only experimental API at the moment.
            /*if error.kind() != ErrorKind::ReadOnlyFilesystem {
                warn!("could not create shutdown PID file: {}", error);
            }*/
            eprintln!("shutdown: warning cannot open {}", error);
        }
        Ok(mut file) => {
            if let Err(error) = file.set_permissions(Permissions::from_mode(0o644)) {
                error!("could not set shutdown PID file permissions: {}", error);
                process::exit(1);
            }
            if let Err(error) = file.write_all(process::id().to_string().as_bytes()) {
                error!("could not write to shutdown PID file: {}", error);
                process::exit(1);
            }
            if let Err(error) = file.sync_all() {
                error!("could not flush shutdown PID file to disk: {}", error);
                process::exit(1);
            }
        }
    }

    // catch some common signals
    unsafe {
        signal::signal(Signal::SIGQUIT, SigHandler::SigIgn).unwrap();
        signal::signal(Signal::SIGCHLD, SigHandler::SigIgn).unwrap();
        signal::signal(Signal::SIGHUP, SigHandler::SigIgn).unwrap();
        signal::signal(Signal::SIGTSTP, SigHandler::SigIgn).unwrap();
        signal::signal(Signal::SIGTTIN, SigHandler::SigIgn).unwrap();
        signal::signal(Signal::SIGTTOU, SigHandler::SigIgn).unwrap();

        let sig_handler = SigHandler::Handler(stop_it);
        let sig_action = SigAction::new(
            sig_handler,
            signal::SaFlags::empty(),
            signal::SigSet::empty(),
        );
        signal::sigaction(Signal::SIGINT, &sig_action).unwrap();
    }

    if args.fast_boot {
        if let Err(error) = File::create(paths::FASTBOOT) {
            error!("FASTBOOT file creation failed: {}", error);
            process::exit(1);
        }

        debug!("enabled FASTBOOT for next boot");
    }
    if args.force_fsck {
        if let Err(error) = File::create(paths::FORCEFSCK) {
            error!("FORCEFSCK file creation failed: {}", error);
            process::exit(1);
        }

        debug!("forced FSCK for next book");
    }

    let mut wait_duration = time_to_duration(&time);
    debug!("we should reboot in {:?}", wait_duration);

    // shutdown NOW if time == 0
    if wait_duration.is_zero() {
        issue_shutdown(
            halt_type,
            down_level,
            args.dont_shut,
            args.no_init,
            args.warn_delay,
            &warn_message,
            new_state,
        );
    }

    /* Rather than loop and reduce [wait_duration] once per minute,
       we shall check the current time against the target time.
       Then calculate the remaining waiting time based on the difference
       between current time and target time.
       This avoids missing shutdown time (target time) after the
       computer has been asleep. -- Jesse
    */
    // target time = current time + wait duration
    let mut local_time = Local::now();
    let target_time = local_time + ChronoDuration::from_std(wait_duration).unwrap();

    // give warnings on regular intervals and finally shutdown
    if wait_duration.as_secs() < 15 * 60
        && !need_warning(wait_duration, args.quiet_full, args.quiet_partial)
    {
        issue_warn(wait_duration, &warn_message, new_state);
    }

    let mut did_nologin = false;
    while local_time < target_time {
        // create the nologin file
        if !did_nologin && wait_duration.as_secs() <= 5 * 60 {
            match File::create(paths::NOLOGIN) {
                Err(error) => warn!("could not create {} file: {}", paths::NOLOGIN, error),
                Ok(mut file) => {
                    let mut message = String::from(&format!(
                        "The system is going down on {}\r\n",
                        (local_time + ChronoDuration::from_std(wait_duration).unwrap())
                            .format("%c")
                    ));

                    // append warning message
                    message.push_str(&warn_message);

                    if let Err(error) = file.write_all(message.as_bytes()) {
                        error!("could not write to {} file: {}", paths::NOLOGIN, error);
                        process::exit(1);
                    }

                    did_nologin = true;
                    debug!("successfully wrote to {} file", paths::NOLOGIN);
                }
            }
        }

        if need_warning(wait_duration, args.quiet_full, args.quiet_partial) {
            issue_warn(wait_duration, &warn_message, new_state);
        }

        // less than 1 min remains : sleep the exact number of seconds
        // more than 1 min remains : sleep 1 minute before looping again
        wait_duration = (target_time - local_time).to_std().unwrap();
        let sleep_duration = wait_duration.min(Duration::from_secs(60));
        debug!("now sleeping for {:?}...", sleep_duration);
        thread::sleep(sleep_duration);

        local_time = Local::now();
    }

    issue_shutdown(
        halt_type,
        down_level,
        args.dont_shut,
        args.no_init,
        args.warn_delay,
        &warn_message,
        new_state,
    );
}
