use std::{fs::File, io::Write as _, iter, time::SystemTime};

use nix::{
    libc,
    sys::utsname::uname,
    unistd::{self, AccessFlags, Pid},
};
use serde::{
    ser::{SerializeStruct as _, Serializer},
    Serialize,
};

use crate::util;

const WTMP_FILE: &str = "/var/log/wtmp";

pub const UT_LINESIZE: usize = 32;
pub const UT_NAMESIZE: usize = 32;
pub const UT_HOSTSIZE: usize = 256;

// from init.c
const THIS_LEVEL: u8 = b'S'; // The current runlevel
const PREV_LEVEL: u8 = b'N'; // Previous runlevel

#[allow(unused)]
#[derive(Debug, PartialEq)]
pub enum RecordType {
    Empty = 0,    // Record does not contain valid info (formerly known as UT_UNKNOWN on Linux)
    RunLvl = 1,   // Change in system run-level (see init(1))
    BootTime = 2, // Time of system boot (in ut_tv)
    NewTime = 3,  // Time after system clock change (in ut_tv)
    OldTime = 4,  // Time before system clock change (in ut_tv)
    InitProcess = 5, // Process spawned by init(1)
    LoginProcess = 6, // Session leader process for user login
    UserProcess = 7, // Normal process
    DeadProcess = 8, // Terminated process
    Accounting = 9, // Not implemented
}

#[repr(C)]
#[derive(Serialize)]
struct ExitStatus {
    e_termination: libc::c_short, // Process termination status
    e_exit: libc::c_short,        // Process exit status
}

struct SerializableTimeval(libc::timeval);
impl Serialize for SerializableTimeval {
    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: Serializer,
    {
        let mut state = serializer.serialize_struct("SerializableTimeval", 2)?;
        state.serialize_field("tv_sec", &self.0.tv_sec)?;
        state.serialize_field("tv_usec", &self.0.tv_usec)?;
        state.end()
    }
}

#[repr(C)]
#[derive(Serialize)]
struct Utmp<'a> {
    ut_type: libc::c_short,                    // Type of record
    ut_pid: libc::pid_t,                       // PID of login process
    ut_line: &'a [libc::c_uchar; UT_LINESIZE], // Device name of tty - "/dev/"
    ut_id: &'a [libc::c_uchar; 4],             // Terminal name suffix, or inittab(5) ID
    ut_user: &'a [libc::c_uchar; UT_NAMESIZE], // Username
    #[serde(serialize_with = "util::serialize_big_arrays")]
    ut_host: &'a [libc::c_uchar; UT_HOSTSIZE], // Hostname for remote login, or kernel version for run-level messages
    ut_exit: ExitStatus, // Exit status of a process marked as DEAD_PROCESS; not used by Linux init(1)
    ut_session: libc::c_long, // Session ID
    ut_tv: SerializableTimeval, // Time entry was made
    ut_addr_v6: &'a [i32; 4], // Internet address of remote host; IPv4 address uses just ut_addr_v6[0]
    __unused: &'a [libc::c_uchar; 20], // Reserved for future use
}

// Log an event in the wtmp file (reboot, runlevel).
pub fn write_wtmp(
    user: &str,              // name of user
    inittab_id: &str,        // inittab ID
    pid: Pid,                // PID of process
    record_type: RecordType, // TYPE of entry
    line: &str,              // Which line is this
) {
    // Can't do much if WTMP_FILE is not present or not writable
    if unistd::access(WTMP_FILE, AccessFlags::W_OK).is_err() {
        warn!("skipping writing to wtmp (missing or not writable)...");
        return;
    }

    // Try to open the wtmp file
    let file = File::options().append(true).create(true).open(WTMP_FILE);
    if let Err(error) = file {
        warn!("could not open {}: {}", WTMP_FILE, error);
        return;
    }

    // Note if we are going to write a boot record
    let wrote_wtmp_reboot = record_type == RecordType::BootTime;

    /*
     *  See if we need to write a reboot record. The reason that
     *  we are being so paranoid is that when we first tried to
     *  write the reboot record, /var was possibly not mounted
     *  yet. As soon as we can open WTMP we write a delayed boot record.
     */
    if !wrote_wtmp_reboot && record_type != RecordType::BootTime {
        write_wtmp("reboot", "~~", Pid::from_raw(0), RecordType::BootTime, "~");
    }

    // Note if we are going to write a runlevel record
    let wrote_wtmp_rlevel = record_type == RecordType::RunLvl;

    /*
     *  See if we need to write a runlevel record. The reason that
     *  we are being so paranoid is that when we first tried to
     *  write the reboot record, /var was possibly not mounted
     *  yet. As soon as we can open WTMP we write a delayed runlevel record.
     */
    if !wrote_wtmp_rlevel && record_type != RecordType::RunLvl {
        let runlevel: u16 = THIS_LEVEL.into();
        let oldlevel: u16 = PREV_LEVEL.into();
        write_wtmp(
            "runlevel",
            "~~",
            Pid::from_raw((runlevel + 256 * oldlevel).into()),
            RecordType::RunLvl,
            "~",
        );
    }

    // Zero the fields and enter new fields.
    let time = SystemTime::now()
        .duration_since(SystemTime::UNIX_EPOCH)
        .expect("clock may have gone backwards");
    let os_release = match uname() {
        Err(error) => {
            warn!(
                "could not retrieve system identification from uname: {}",
                error
            );
            String::from("")
        }
        Ok(utsname) => {
            debug!(
                "retrieved system identification from uname: {:?}",
                utsname.release()
            );
            utsname.release().to_str().unwrap().to_string()
        }
    };

    let utmp = Utmp {
        ut_type: record_type as libc::c_short,
        ut_pid: pid.into(),
        ut_line: &line
            .bytes()
            .chain(iter::repeat(b'\x00'))
            .take(UT_LINESIZE)
            .collect::<Vec<_>>()
            .try_into()
            .unwrap(),
        ut_id: &inittab_id
            .bytes()
            .take(4)
            .collect::<Vec<_>>()
            .try_into()
            .unwrap(),
        ut_user: &user
            .bytes()
            .take(UT_NAMESIZE)
            .collect::<Vec<_>>()
            .try_into()
            .unwrap(),
        ut_host: &os_release
            .bytes()
            .take(UT_HOSTSIZE)
            .collect::<Vec<_>>()
            .try_into()
            .unwrap(),
        // below fields are zeroed in official implementation
        ut_exit: ExitStatus {
            e_termination: 0,
            e_exit: 0,
        },
        ut_session: 0,
        ut_tv: SerializableTimeval(libc::timeval {
            tv_sec: time.as_secs() as libc::time_t,
            tv_usec: time.subsec_micros() as libc::suseconds_t,
        }),
        ut_addr_v6: &vec![0; 4].try_into().unwrap(),
        __unused: &vec![b'\x00'; 20].try_into().unwrap(),
    };

    debug!("writing {} wtmp record to {}...", user, WTMP_FILE);

    if let Err(error) = file.unwrap().write_all(&bincode::serialize(&utmp).unwrap()) {
        warn!("could not write record to wtmp: {}", error);
    }
}
